use crate::util::ToLitStr;
use syn::{spanned::Spanned as _, DataEnum, DataStruct, DataUnion, Fields, FieldsNamed, LitStr, Type};

pub(super) enum ParseDataType {
	Type(Type),
	Inline(ParseData)
}

#[allow(dead_code)]
pub(super) enum ParseData {
	Struct(Vec<(LitStr, ParseDataType)>),
	Enum(Vec<LitStr>),
	Alternatives(Vec<ParseData>),
	Unit
}

fn parse_named_fields(named_fields: &FieldsNamed) -> syn::Result<ParseData> {
	let mut fields: Vec<(LitStr, ParseDataType)> = Vec::new();
	for f in &named_fields.named {
		let ident = f
			.ident
			.as_ref()
			.ok_or_else(|| syn::Error::new(f.span(), "#[derive(OpenapiType)] does not support fields without an ident"))?;
		let name = ident.to_lit_str();
		let ty = f.ty.to_owned();
		fields.push((name, ParseDataType::Type(ty)));
	}
	Ok(ParseData::Struct(fields))
}

pub(super) fn parse_struct(strukt: &DataStruct) -> syn::Result<ParseData> {
	match &strukt.fields {
		Fields::Named(named_fields) => parse_named_fields(named_fields),
		Fields::Unnamed(_) => unimplemented!(),
		Fields::Unit => Ok(ParseData::Unit)
	}
}

pub(super) fn parse_enum(inum: &DataEnum) -> syn::Result<ParseData> {
	let mut strings: Vec<LitStr> = Vec::new();
	let mut types: Vec<(LitStr, ParseData)> = Vec::new();

	for v in &inum.variants {
		let name = v.ident.to_lit_str();
		match &v.fields {
			Fields::Named(named_fields) => {
				types.push((name, parse_named_fields(named_fields)?));
			},
			Fields::Unnamed(_unnamed_fields) => unimplemented!(),
			Fields::Unit => strings.push(name)
		}
	}

	let data_strings = if strings.is_empty() {
		None
	} else {
		Some(ParseData::Enum(strings))
	};

	let data_types = if types.is_empty() {
		None
	} else {
		Some(ParseData::Alternatives(
			types
				.into_iter()
				.map(|(name, data)| ParseData::Struct(vec![(name, ParseDataType::Inline(data))]))
				.collect()
		))
	};

	match (data_strings, data_types) {
		// only variants without fields
		(Some(data), None) => Ok(data),
		// only one variant with fields
		(None, Some(ParseData::Alternatives(mut alt))) if alt.len() == 1 => Ok(alt.remove(0)),
		// only variants with fields
		(None, Some(data)) => Ok(data),
		// variants with and without fields
		(Some(data), Some(ParseData::Alternatives(mut alt))) => {
			alt.push(data);
			Ok(ParseData::Alternatives(alt))
		},
		// no variants
		(None, None) => Err(syn::Error::new(
			inum.brace_token.span,
			"#[derive(OpenapiType)] does not support enums with no variants"
		)),
		// data_types always produces Alternatives
		_ => unreachable!()
	}
}

pub(super) fn parse_union(union: &DataUnion) -> syn::Result<ParseData> {
	Err(syn::Error::new(
		union.union_token.span(),
		"#[derive(OpenapiType)] cannot be used on unions"
	))
}
